from typing import List, Union
from . import (
    germeval2018,
    germeval2017,
    gnad10,
    german_ner,
    german_europarl,
    german_quad,
    spanish_quad,
    wiki_cat_es,
    spanish_conll,
    flue,
    spanish_ehealth,
    szeged_ner,
    polish_dyk,
    mapa,
    eur_lux,
    ehealth_kd,
    rucola,
    klej_dyk,
    croatian_sentiment,
    finish_sentiment,
    swedish_ner,
    greek_legal,
    bulgarian_sentiment,
    czech_subjectivity,
    danish_misogyny,
    slovak_sentiment,
    maltese_sentiment,
    dutch_social,
    classification,
    ner,
    qa,
    
)

from .task_registry import get_all_tasks, TASK_REGISTRY, ALL_TASK_TYPES
